#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of the VPM Project

"""
Toolchain-simplez are a set of tools for working with the education processor
Simplez. The included tools are:

* sasm: Simplez Assembler
* ssim: Simplez simulator
* sreset: Simplez Reset
* sboot: Simplez loader (a bootloader is needed)
"""

__author__ = 'Juan Gonzalez-Gomez (obijuan)'
__email__ = 'juan.gonzalez@bq.com'
__license__ = 'GPLv2'
__version__ = '0.0.11'
